--
-- newSprayer
-- Specialization for newSprayer
--
-- monteur1

newSprayer = {};

function newSprayer.initSpecialization()
    WorkArea.registerAreaType("sprayer");
end

function newSprayer.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Fillable, specializations) and
           SpecializationUtil.hasSpecialization(WorkArea, specializations) and
           SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations) and
           SpecializationUtil.hasSpecialization(AttacherJoints, specializations);
end

function newSprayer:preLoad(savegame)
	self.getIsReadyToSpray = newSprayer.getIsReadyToSpray;
	self.getAreEffectsVisible = newSprayer.getAreEffectsVisible;
	self.getnewSprayerLitersPerSecond = newSprayer.getnewSprayerLitersPerSecond;
	self.supportsFillTriggers = true;
	self.loadWorkAreaFromXML = Utils.overwrittenFunction(self.loadWorkAreaFromXML, newSprayer.loadWorkAreaFromXML);
	self.doCheckSpeedLimit = Utils.overwrittenFunction(self.doCheckSpeedLimit, newSprayer.doCheckSpeedLimit);
	self.getIsTurnedOnAllowed = Utils.overwrittenFunction(self.getIsTurnedOnAllowed, newSprayer.getIsTurnedOnAllowed);
	self.getCanAddHelpButtonText = Utils.overwrittenFunction(self.getCanAddHelpButtonText, newSprayer.getCanAddHelpButtonText);
	self.getDrawFirstFillText = Utils.overwrittenFunction(self.getDrawFirstFillText, newSprayer.getDrawFirstFillText);
	
	self.aiAllowedToRefill = Utils.overwrittenFunction(self.aiAllowedToRefill, newSprayer.aiAllowedToRefill);
	self.getnewSprayerHasSpray = newSprayer.getnewSprayerHasSpray;
	self.processSprayerAreas = newSprayer.processSprayerAreas;
end

function newSprayer:load(xmlFile)
    self.isSprayerTank = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#isTank"), false);
    self.newSprayerSpecisDeactivated = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#newSprayerSpecisDeactivated"), false);
	
	self.newSprayerFrontDock = {}
    self.isSprayerFrontDock = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#isFrontDock"), false);	

    self.allowsSpraying = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#allowsSpraying"), true);
    self.needsTankActivation = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#needsTankActivation"), false);
    self.activateTankOnLowering = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#activateTankOnLowering"), false);
    self.stopAiOnEmpty = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#stopAiOnEmpty"), true);
    self.activateOnLowering = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#activateOnLowering"), false);
    self.sprayUsageScale = {}
    self.sprayUsageScalenewSprayer = {}
    self.sprayUsageScale.default = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.sprayUsageScales#scale"), 1)
    self.sprayUsageScalenewSprayer.workingWidth = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.sprayUsageScales#workingWidth"), 12)
    self.sprayUsageScale.fillTypeScales = {}
    local i=0;
    while true do
        local key = string.format("vehicle.sprayUsageScales.sprayUsageScale(%d)", i);
        if not hasXMLProperty(self.xmlFile, key) then
            break;
        end
        local fillType = getXMLString(self.xmlFile, key.. "#fillType");
        local scale = getXMLFloat(self.xmlFile, key.. "#scale");
        if fillType ~= nil and scale ~= nil then
            local fillTypeInt = FillUtil.fillTypeNameToInt[fillType];
            if fillTypeInt ~= nil then
                self.sprayUsageScale.fillTypeScales[fillTypeInt] = scale
            else
                print("Warning: Invalid spray usage scale fill type '"..fillType.."' in '" .. self.configFileName.. "'");
            end
        end
        i = i+1;
    end
    self.lastSprayValveUpdateFoldTime = nil;
    self.sprayParticleSystems = {};
    self.newsprayerEffects = {};
    if self.isClient then
        local i = 0;
        while true do
            local key = string.format("vehicle.sprayParticleSystems.sprayParticleSystem(%d)", i);
            local t = getXMLString(self.xmlFile, key .. "#type");
            if t == nil then
                break;
            end
            local fillType = FillUtil.fillTypeNameToInt[t];
            if fillType ~= nil then
                self.sprayParticleSystems[fillType] = {}
                local currentPS = Utils.getNoNil(self.sprayParticleSystems[fillType], {});
                ParticleUtil.loadParticleSystem(self.xmlFile, currentPS, key, self.components, false, nil, self.baseDirectory);
                currentPS.foldMinLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#foldMinLimit"), 0);
                currentPS.foldMaxLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#foldMaxLimit"), 1);
                table.insert(self.sprayParticleSystems[fillType], currentPS)
            end
            i = i + 1;
        end
        local i = 0
        while true do
            local key = string.format("vehicle.sprayParticles.emitterShape(%d)", i);
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end
            local emitterShape = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#node"));
            local particleType = getXMLString(self.xmlFile, key.."#particleType")
            if emitterShape ~= nil then
                for index, _ in pairs(self.fillUnits) do
                    for fillType, _ in pairs(self:getUnitFillTypes(index)) do
                        if self.sprayParticleSystems[fillType] == nil then
                            self.sprayParticleSystems[fillType] = {}
                        end
                        local particleSystem = MaterialUtil.getParticleSystem(fillType, particleType)
                        if particleSystem ~= nil then
                            local currentPS = ParticleUtil.copyParticleSystem(self.xmlFile, key, particleSystem, emitterShape)
                            currentPS.foldMinLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#foldMinLimit"), 0);
                            currentPS.foldMaxLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#foldMaxLimit"), 1);
                            table.insert(self.sprayParticleSystems[fillType], currentPS)
                        end
                    end
                end
            end
            i = i + 1
        end
        local i = 0;
        while true do
            local key = string.format("vehicle.sprayerEffects.sprayerEffect(%d)", i);
            if not hasXMLProperty(self.xmlFile, key) then
                break;
            end
            local effects = EffectManager:loadEffect(self.xmlFile, key, self.components, self);
            if effects ~= nil then
                local sprayerEffect = {}
                sprayerEffect.effects = effects
                sprayerEffect.foldMinLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, key .. "#foldMinLimit"), 0);
                sprayerEffect.foldMaxLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, key .. "#foldMaxLimit"), 1);
                table.insert(self.newsprayerEffects, sprayerEffect);
            end
            i = i + 1;
        end
        self.sprayingAnimationName = Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.sprayingAnimation#name"), "");
        self.sampleSprayer = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.spraySound", nil, self.baseDirectory);
        if self.sampleFill == nil then
            local linkNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.fillSound#linkNode"), "0>"));
            self.sampleFill = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.fillSound", nil, self.baseDirectory, linkNode);
        end
        self.sampleFillEnabled = false;
        self.sampleFillStopTime = -1;
        self.lastFillLevel = -1;
        self.newsprayerTurnedOnRotationNodes = Utils.loadRotationNodes(self.xmlFile, {}, "vehicle.turnedOnRotationNodes.turnedOnRotationNode", "sprayer", self.components);
        self.newsprayerTurnedOnScrollers = Utils.loadScrollers(self.components, self.xmlFile, "vehicle.sprayerTurnedOnScrollers.sprayerTurnedOnScroller", {}, false);
    end
    self.showFieldNotOwnedWarning = false;
    self.isSprayerSpeedLimitActive = false;
    if self.sowingMachineGroundContactFlag == nil and self.cultivatorGroundContactFlag == nil then
        table.insert(self.terrainDetailRequiredValueRanges, {g_currentMission.ploughValue, g_currentMission.ploughValue, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels});
        table.insert(self.terrainDetailRequiredValueRanges, {g_currentMission.cultivatorValue, g_currentMission.cultivatorValue, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels});
        table.insert(self.terrainDetailRequiredValueRanges, {g_currentMission.sowingValue, g_currentMission.sowingValue, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels});
        table.insert(self.terrainDetailRequiredValueRanges, {g_currentMission.sowingWidthValue, g_currentMission.sowingWidthValue, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels});
        table.insert(self.terrainDetailRequiredValueRanges, {g_currentMission.grassValue, g_currentMission.grassValue, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels});
        table.insert(self.terrainDetailProhibitValueRanges, {1, 2, g_currentMission.sprayFirstChannel, g_currentMission.sprayNumChannels});
        table.insert(self.terrainDetailProhibitValueRanges, {g_currentMission.sprayLevelMaxValue, g_currentMission.sprayLevelMaxValue, g_currentMission.sprayLevelFirstChannel, g_currentMission.sprayLevelNumChannels});
    end
    self.newsprayer = {};
    self.newsprayer.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sprayer#fillUnitIndex"), 1);
    self.newsprayer.unloadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sprayer#unloadInfoIndex"), 1);
    self.newsprayer.loadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sprayer#loadInfoIndex"), 1);
    self.newsprayer.dischargeInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sprayer#dischargeInfoIndex"), 1);
    self.newsprayerDirtyFlag = self:getNextDirtyFlag();
	
	self.newsprayerIsRealSprayer = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#isRealSprayer"), false);
	
	self.newsprayAppRateSoll = 30 -- 30 Kubikmeter Guelle oder 500 Liter Spritzmittel (multiplyer 16.666) pro Hektar bei 10km/h 
	
	if self.newsprayerIsRealSprayer == false then
		self.newsprayAppRateMultiplyer = self.newsprayAppRateSoll / 11.7  -- 0 = 11.7 kubikmeter Guelle pro Hektar bei 10km/h 
	else
		self.newsprayAppRateMultiplyer = self.newsprayAppRateSoll / 11.7  -- 0 = 11.7 liter Spritzmittel pro Hektar bei 10km/h 
	end
	
	self.newsprayerWidth = 1;
	self.newsprayerTextureDisable = false
	
	self.newsprayerHUD = Overlay:new("Sprayeroverlay", Utils.getFilename("SprayerARHud.dds", self.baseDirectory.."overlays/"),  0.82, 0.22, 0.10, 0.04);
	self.newsprayerAppRate = 0;
	self.newsprayerAktWorkSpeed = 0
end

function newSprayer:delete()
    for _, sprayerEffect in pairs(self.newsprayerEffects) do
        EffectManager:deleteEffects(sprayerEffect.effects);
    end
    for _, particleSystems in pairs(self.sprayParticleSystems) do
        ParticleUtil.deleteParticleSystems(particleSystems)
    end
    if self.isClient then
        SoundUtil.deleteSample(self.sampleSprayer);
        SoundUtil.deleteSample(self.sampleFill);
    end
end

function newSprayer:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        self.showFieldNotOwnedWarning = streamReadBool(streamId);		
    end
end

function newSprayer:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        streamWriteBool(streamId, self.showFieldNotOwnedWarning);
    end
end

function newSprayer:readStream(streamId, connection)
	self.newsprayAppRateSoll = streamReadFloat32(streamId);
	self.newsprayAppRateMultiplyer = streamReadFloat32(streamId);
	self.newsprayerAppRate = streamReadFloat32(streamId);		
	self.newsprayerWidth = streamReadFloat32(streamId);	
	self.newsprayerTextureDisable = streamReadBool(streamId);			
end

function newSprayer:writeStream(streamId, connection)		
	streamWriteFloat32(streamId, self.newsprayAppRateSoll);
	streamWriteFloat32(streamId, self.newsprayAppRateMultiplyer);
	streamWriteFloat32(streamId, self.newsprayerAppRate);		
	streamWriteFloat32(streamId, self.newsprayerWidth);	
	streamWriteBool(streamId, self.newsprayerTextureDisable);		
end

function newSprayer:update(dt)
    if self.isClient and self.newSprayerSpecisDeactivated == false then
        Utils.updateRotationNodes(self, self.newsprayerTurnedOnRotationNodes, dt, self:getIsActive() and self:getIsTurnedOn());
        Utils.updateScrollers(self.newsprayerTurnedOnScrollers, dt, self:getIsActive() and self:getIsTurnedOn());
    end

	if self:getIsActive() and self.newSprayerSpecisDeactivated == false then
		if self.isClient and self:getIsActiveForInput(true) and self.isSprayerTank == false then	
			local aktualSprayrate = self.newsprayAppRateSoll
			local AppRateMultiplyer = self.newsprayAppRateMultiplyer

			if InputBinding.hasEvent(InputBinding.INCREASE_SPRAYRATE) then  --Ausbringmenge erhöhen
				local newSprayrate = aktualSprayrate + 0.5
				if newSprayrate > 40 then
					newSprayrate = 40
				end
				self.newsprayAppRateSoll = newSprayrate
				AppRateMultiplyer = newSprayrate / 11.7	-- 0 = 11.7 liter pro Hektar bei 10km/h 
				self.newsprayAppRateMultiplyer = AppRateMultiplyer
				if g_server ~= nil then
					g_server:broadcastEvent(newSprayer_Sprayrate_Event:new(self, self.newsprayAppRateMultiplyer), nil, nil, self);
				else
					g_client:getServerConnection():sendEvent(newSprayer_Sprayrate_Event:new(self, self.newsprayAppRateMultiplyer));
				end;							
			end
		
			if InputBinding.hasEvent(InputBinding.DECREASE_SPRAYRATE) then  --Ausbringmenge reduzieren
				local newSprayrate = aktualSprayrate - 0.5
				if newSprayrate < 10 then
					newSprayrate = 10
				end			
				self.newsprayAppRateSoll = newSprayrate
				AppRateMultiplyer = newSprayrate / 11.7	-- 0 = 11.7 liter pro Hektar bei 10km/h 
				self.newsprayAppRateMultiplyer = AppRateMultiplyer	
				if g_server ~= nil then
					g_server:broadcastEvent(newSprayer_Sprayrate_Event:new(self, self.newsprayAppRateMultiplyer), nil, nil, self);
				else
					g_client:getServerConnection():sendEvent(newSprayer_Sprayrate_Event:new(self, self.newsprayAppRateMultiplyer));
				end;							
			end
			
			if InputBinding.hasEvent(InputBinding.TEXTURE_DISABLE) then  --Gülletextur an/aus
				if self.newsprayerTextureDisable == true then
					self.newsprayerTextureDisable = false
					if g_server ~= nil then
						g_server:broadcastEvent(newSprayer_TextureDisable_Event:new(self, self.newsprayerTextureDisable), nil, nil, self);
					else
						g_client:getServerConnection():sendEvent(newSprayer_TextureDisable_Event:new(self, self.newsprayerTextureDisable));
					end;												
				elseif self.newsprayerTextureDisable == false then
					self.newsprayerTextureDisable = true
					if g_server ~= nil then
						g_server:broadcastEvent(newSprayer_TextureDisable_Event:new(self, self.newsprayerTextureDisable), nil, nil, self);
					else
						g_client:getServerConnection():sendEvent(newSprayer_TextureDisable_Event:new(self, self.newsprayerTextureDisable));
					end;																	
				end
			end				
		end				
	end
end

function newSprayer:updateTick(dt)
    self.isSprayerSpeedLimitActive = false;
    if self:getIsActive() and self.newSprayerSpecisDeactivated == false then
        local showFieldNotOwnedWarning = false;
        if self.isServer then
            if self.turnOnDueToLoweredImplement ~= nil then
                if self.turnOnDueToLoweredImplement.object == nil then
                    self.turnOnDueToLoweredImplement = nil;
                end
            end
            if self.turnOnDueToLoweredImplement ~= nil then
                if self:getIsTurnedOn() == false then
                    self:setIsTurnedOn(true);
                end
                if self.turnOnDueToLoweredImplement.object.activateOnLowering then
                    if self.turnOnDueToLoweredImplement.object:getIsTurnedOn() == false then
                        self.turnOnDueToLoweredImplement.object:setIsTurnedOn(true);
                    end
                end
            end
            if self:getIsTurnedOn() then
                if not self:getIsTurnedOnAllowed(true) then
                    self:setIsTurnedOn(false);
                end
            end
            local tankActivators = {};
            newSprayer.getTankActivators(self:getRootAttacherVehicle(), tankActivators);
            if table.getn(tankActivators) > 0 then
                local doTurnOn = false;
                for _,tankActivator in pairs(tankActivators) do
                    local attacherVehicle = tankActivator.attacherVehicle;
                    if attacherVehicle ~= nil then
                        local implementIndex = attacherVehicle:getImplementIndexByObject(tankActivator);
                        local implement = attacherVehicle.attachedImplements[implementIndex];
                        local jointDescIndex = implement.jointDescIndex;
                        doTurnOn = doTurnOn or attacherVehicle.attacherJoints[jointDescIndex].moveDown;
                    end
                end
                self:setIsTurnedOn(doTurnOn);
            end
            if self.activateOnLowering and self.isLowered ~= nil then
                if self:isLowered(false) then
                    if self:getCanBeTurnedOn() and not self:getIsTurnedOn(true) then
                        self:setIsTurnedOn(true);
                    end
                else
                    self:setIsTurnedOn(false);
                end
            end
        end
        if self:getIsTurnedOn() and self.allowsSpraying then
            if self.isClient and self:getIsActiveForSound() then
                SoundUtil.playSample(self.sampleSprayer, 0, 0, nil);
            end
            local readyToSpray = self:getIsReadyToSpray();
            self.isSprayerSpeedLimitActive = readyToSpray;
            local fillType = self:getUnitLastValidFillType(self.newsprayer.fillUnitIndex);
            if fillType == FillUtil.FILLTYPE_UNKNOWN and self.fillUnits[self.newsprayer.fillUnitIndex] ~= nil then
                for unitFillType,state in pairs(self.fillUnits[self.newsprayer.fillUnitIndex].fillTypes) do
                    if unitFillType ~= FillUtil.FILLTYPE_UNKNOWN and state then
                        fillType = unitFillType;
                        break;
                    end
                end
            end
			
            if readyToSpray then
			
---- Sprayer Application Rate calc Workingwidth-----------------------------	
				local xmin = 200;
				local xmax = -200;
				for _,workAreas in pairs(self.workAreas) do							
					local x1,y1,z1 = getWorldTranslation(workAreas.start)
					local x2,y2,z2 = getWorldTranslation(workAreas.width)
					local x3,y3,z3 = getWorldTranslation(workAreas.height)
					local lx1,ly1,lz1 = worldToLocal(self.rootNode,x1,y1,z1)
					local lx2,ly2,lz2 = worldToLocal(self.rootNode,x2,y2,z2)
					local lx3,ly3,lz3 = worldToLocal(self.rootNode,x3,y3,z3)
			
					if lx1 < xmin then
						xmin = lx1;
					end
					if lx1 > xmax then
						xmax = lx1;
					end
					if lx2 < xmin then
						xmin = lx2;
					end
					if lx2 > xmax then
						xmax = lx2;
					end
					if lx3 < xmin then
						xmin = lx3;
					end
					if lx3 > xmax then
						xmax = lx3;
					end
				end			
		
				self.newsprayerWidth = xmax-xmin;
		
				if self.newsprayerWidth < 0 then
					self.newsprayerWidth = 0;
				end;
	
				local spraywid = self.newsprayerWidth
				self.sprayUsageScalenewSprayer.workingWidth	= self.newsprayerWidth
---- Sprayer Application Rate calc Ende -----------------------------	
			
                if self.isServer then									
                    local litersPerSecond = self:getnewSprayerLitersPerSecond(fillType);
                    local usage = litersPerSecond * dt * 0.001;
                    local hasSpray, newFillType = self:getnewSprayerHasSpray(fillType, usage)					
                    fillType = newFillType
                    if hasSpray then
                        local workAreas, showWarning, _ = self:getTypedNetworkAreas(WorkArea.AREATYPE_SPRAYER, true);
                        showFieldNotOwnedWarning = showWarning;
                        if (table.getn(workAreas) > 0) then
                            local pixels = self:processSprayerAreas(workAreas, fillType);
                            local ha = Utils.areaToHa(pixels, g_currentMission:getFruitPixelsToSqm());
                            g_currentMission.missionStats:updateStats("fertilizedHectares", ha);
                            g_currentMission.missionStats:updateStats("fertilizedTime", dt/(1000*60));
                            if self.lastSowingArea == nil then
                                g_currentMission.missionStats:updateStats("workedHectares", ha);
                                g_currentMission.missionStats:updateStats("workedTime", dt/(1000*60));
                            end
                            g_currentMission.missionStats:updateStats("sprayUsage", usage);
                        end
					else
						self:setIsTurnedOn(false)					
                    end								
                end
							
---- Sprayer Application Rate -----------------------------	
				if self.isClient then
					local kmh = 0;					
					if self.attacherVehicle ~= nil then
						if self.attacherVehicle.lastSpeed ~= nil and self.attacherVehicle.speedDisplayScale ~= nil then
							kmh = self.attacherVehicle.lastSpeed*self.attacherVehicle.speedDisplayScale*3600;
							if kmh < 0.1 then
								kmh = 0;
							end;
							self.newsprayerAktWorkSpeed = kmh
						end;
					end;				
					self.newsprayerAppRate = 0.0;
                    local litersPerSecond = self:getnewSprayerLitersPerSecond(fillType);
					self.newsprayerAppRate = (litersPerSecond*36.0/(kmh*self.newsprayerWidth));	
				end				
---- Sprayer Application Rate Ende -----------------------------					
				
            end
            if self.isClient then			
                -- update valve particle systems
                local foldAnimTime = self.foldAnimTime				
				if (foldAnimTime ~= nil and foldAnimTime ~= self.lastSprayValveUpdateFoldTime) or self.lastTurnedOn == false then
					self.lastSprayValveUpdateFoldTime = foldAnimTime;
					self.lastTurnedOn = true;
					if self:getAreEffectsVisible() then
						if foldAnimTime ~= nil then
							if self.sprayParticleSystems[fillType] ~= nil then
								for _, ps in pairs(self.sprayParticleSystems[fillType]) do
									if foldAnimTime <= ps.foldMaxLimit and foldAnimTime >= ps.foldMinLimit then
										ParticleUtil.setEmittingState(ps, true);
									else
										ParticleUtil.setEmittingState(ps, false);
									end
								end;
							end
							for _, sprayerEffect in pairs(self.newsprayerEffects) do
								readyToSpray = readyToSpray and (foldAnimTime > sprayerEffect.foldMinLimit);
								if foldAnimTime <= sprayerEffect.foldMaxLimit and foldAnimTime >= sprayerEffect.foldMinLimit then
									EffectManager:setFillType(sprayerEffect.effects, fillType)
									EffectManager:startEffects(sprayerEffect.effects)
								else
									EffectManager:stopEffects(sprayerEffect.effects);
								end
							end
						else
							if self.sprayParticleSystems[fillType] ~= nil then
								for _, ps in pairs(self.sprayParticleSystems[fillType]) do
									ParticleUtil.setEmittingState(ps, true);
								end;
							end
							for _, sprayerEffect in pairs(self.newsprayerEffects) do
								EffectManager:setFillType(sprayerEffect.effects, fillType)
								EffectManager:startEffects(sprayerEffect.effects)
							end
						end
					end
				end
            end
        else
            if self.isClient then
                for _, particleSystems in pairs(self.sprayParticleSystems) do
                    for _, ps in pairs(particleSystems) do
                        ParticleUtil.setEmittingState(ps, false);
                    end;
                end
                for _, sprayerEffect in pairs(self.newsprayerEffects) do
                    EffectManager:stopEffects(sprayerEffect.effects);
                end
            end
        end
        self.lastTurnedOn = self:getIsTurnedOn();
        if self.isServer then
            if self.sowingMachineGroundContactFlag == nil then
                if showFieldNotOwnedWarning ~= self.showFieldNotOwnedWarning then
                    self.showFieldNotOwnedWarning = showFieldNotOwnedWarning;
                    self:raiseDirtyFlags(self.newsprayerDirtyFlag);
                end
            end
        end
    end
    if self.isClient and self.newSprayerSpecisDeactivated == false then
        if self.isFilling then
            if self:getIsActiveForSound(true) then
                SoundUtil.playSample(self.sampleFill, 0, 0, nil);
                SoundUtil.stop3DSample(self.sampleFill);
            else
                SoundUtil.stopSample(self.sampleFill);
                SoundUtil.play3DSample(self.sampleFill);
            end
        else
            SoundUtil.stopSample(self.sampleFill);
            SoundUtil.stop3DSample(self.sampleFill);
        end
    end
end

function newSprayer:draw()
    if self.isClient and self.newSprayerSpecisDeactivated == false then
		if self.isSprayerTank == false then
			if self.showFieldNotOwnedWarning then
				g_currentMission:showBlinkingWarning(g_i18n:getText("warning_youDontOwnThisField"));
			end
			if self:getIsActiveForInput(true) then	
				local multiplyer = 1
				if self.newsprayerIsRealSprayer == true then
					multiplyer = 16.666666
				end
				
				self.newsprayerHUD:render();
				setTextAlignment(RenderText.ALIGN_RIGHT); 
				setTextBold(true);
				local r = .8;
				local g = .8;
				local b = .8;
				if self.newsprayerAppRate > 0 then
					r = 0.8;
					g = 0;
					b = 0;
			
					if (self.newsprayerAppRate*multiplyer) > 6 and (self.newsprayerAppRate*multiplyer) < 40 then
						r = 0.8;
						g = .8;
						b = 0;
					end
					if (self.newsprayerAppRate*multiplyer) > 9 and (self.newsprayerAppRate*multiplyer) < 35 then
						r = 0;
						g = .8;
						b = 0;
					end
				end
								
				setTextColor(r,g,b,1);
				if self.newsprayerAppRate ~= nil then
					renderText(.82 + 0.085,0.22 + 0.010,0.022, string.format("%3.1f",(self.newsprayerAppRate*multiplyer)));
				end
				
				setTextAlignment(RenderText.ALIGN_LEFT); 
				setTextBold(false);	
				setTextColor(1,1,1,1);
			
				renderText(.82 + 0.03,0.22 + 0.010,0.022, string.format("%3.1f",(self.newsprayerAppRate*multiplyer)));
			
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("input_INCREASE_SPRAYRATE")), InputBinding.INCREASE_SPRAYRATE, nil, GS_PRIO_HIGH)					
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("input_DECREASE_SPRAYRATE")), InputBinding.DECREASE_SPRAYRATE, nil, GS_PRIO_HIGH)	
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("input_TEXTURE_DISABLE")), InputBinding.TEXTURE_DISABLE, nil, GS_PRIO_HIGH)	
			end
		end
	end
end

function newSprayer:getDrawFirstFillText(superFunc)
    if self.isClient then
        if self:getIsActiveForInput(true) and not self.isAlwaysTurnedOn then
            if not self:getIsTurnedOnAllowed(true) and self:getUnitFillLevel(self.newsprayer.fillUnitIndex) <= 0 and self:getUnitCapacity(self.newsprayer.fillUnitIndex) > 0 then
                return true;
            end
        end
    end
    if superFunc ~= nil then
        return superFunc(self);
    end;
    return false;
end;

function newSprayer:onDetach(attacherVehicle, jointDescIndex)
    if attacherVehicle.setIsTurnedOn ~= nil and attacherVehicle:getIsTurnedOn() then
        attacherVehicle:setIsTurnedOn(false);
    end
end

function newSprayer:onDeactivate()
    self.showFieldNotOwnedWarning = false;
end

function newSprayer:onDeactivateSounds()
    if self.isClient then
        SoundUtil.stopSample(self.sampleSprayer, true);
    end
end

function newSprayer:getnewSprayerHasSpray(fillType, usage)
    local hasSpray = false;
    if not hasSpray and self:getIsHired() and self:aiAllowedToRefill() then
        if fillType == FillUtil.FILLTYPE_LIQUIDMANURE or
           fillType == FillUtil.FILLTYPE_DIGESTATE or
           (fillType == FillUtil.FILLTYPE_UNKNOWN and (self:allowUnitFillType(self.newsprayer.fillUnitIndex, FillUtil.FILLTYPE_LIQUIDMANURE) or self:allowUnitFillType(self.newsprayer.fillUnitIndex, FillUtil.FILLTYPE_DIGESTATE))) then
            if g_currentMission.missionInfo.helperSlurrySource == 2 then -- buy manure
                hasSpray = true;
                if g_currentMission.economyManager:getCostPerLiter(FillUtil.FILLTYPE_LIQUIDMANURE) then
                    fillType = FillUtil.FILLTYPE_LIQUIDMANURE;
                else
                    fillType = FillUtil.FILLTYPE_DIGESTATE;
                end
                local price = usage * g_currentMission.economyManager:getCostPerLiter(fillType) * 1.5  -- increase price if AI is active to reward the player's manual work
                g_currentMission.missionStats:updateStats("expenses", price);
                g_currentMission:addSharedMoney(-price, "purchaseFertilizer");
            elseif g_currentMission.missionInfo.helperSlurrySource > 2 then
                local trigger = g_currentMission.liquidManureTriggers[g_currentMission.missionInfo.helperSlurrySource-2].trigger
                if trigger.fillLevel > 0 then
                    hasSpray = true;
                    fillType = trigger.fillType
                    trigger:setFillLevel(trigger.fillLevel-usage)
                end
            end
        elseif fillType == FillUtil.FILLTYPE_MANURE or (fillType == FillUtil.FILLTYPE_UNKNOWN and self:allowUnitFillType(self.newsprayer.fillUnitIndex, FillUtil.FILLTYPE_MANURE)) then
            if g_currentMission.missionInfo.helperManureSource == 2 then -- buy manure
                hasSpray = true;
                fillType = FillUtil.FILLTYPE_MANURE;
                local price = usage * g_currentMission.economyManager:getCostPerLiter(fillType) * 1.5  -- increase price if AI is active to reward the player's manual work
                g_currentMission.missionStats:updateStats("expenses", price);
                g_currentMission:addSharedMoney(-price, "purchaseFertilizer");
            elseif g_currentMission.missionInfo.helperManureSource > 2 then
                local manureHeap = g_currentMission.manureHeaps[g_currentMission.missionInfo.helperManureSource-2].manureHeap
                if manureHeap:removeManure(usage) > 0 then
                    hasSpray = true;
                    fillType = FillUtil.FILLTYPE_MANURE
                end
            end
        elseif fillType == FillUtil.FILLTYPE_FERTILIZER or
               fillType == FillUtil.FILLTYPE_LIQUIDFERTILIZER or
              (fillType == FillUtil.FILLTYPE_UNKNOWN and (self:allowUnitFillType(self.newsprayer.fillUnitIndex, FillUtil.FILLTYPE_LIQUIDFERTILIZER) or self:allowUnitFillType(self.newsprayer.fillUnitIndex, FillUtil.FILLTYPE_FERTILIZER))) then
            if g_currentMission.missionInfo.helperBuyFertilizer then
                hasSpray = true;
                if self:allowUnitFillType(self.newsprayer.fillUnitIndex, FillUtil.FILLTYPE_LIQUIDFERTILIZER) then
                    fillType = FillUtil.FILLTYPE_LIQUIDFERTILIZER;
                else
                    fillType = FillUtil.FILLTYPE_FERTILIZER;
                end
                local price = usage * g_currentMission.economyManager:getCostPerLiter(fillType) * 1.5  -- increase price if AI is active to reward the player's manual work
                g_currentMission.missionStats:updateStats("expenses", price);
                g_currentMission:addSharedMoney(-price, "purchaseFertilizer");
            end
        else
        end
    end
    if not hasSpray then
        local sprayUsageVehicle = self
        local newFillType = fillType;
        if self:getUnitFillLevel(self.newsprayer.fillUnitIndex) == 0 then
            sprayUsageVehicle, newFillType = newSprayer.findAttachedSprayerTank(self:getRootAttacherVehicle(), self:getUnitFillTypes(self.newsprayer.fillUnitIndex), self.needsTankActivation);
        end
        local oldFillLevel = 0
		local newFillLevel = 0
        if sprayUsageVehicle ~= nil then
            fillType = newFillType;
            oldFillLevel = sprayUsageVehicle:getUnitFillLevel(sprayUsageVehicle.newsprayer.fillUnitIndex);
			newFillLevel = oldFillLevel - usage
			usage = usage
            sprayUsageVehicle:setUnitFillLevel(sprayUsageVehicle.newsprayer.fillUnitIndex, newFillLevel, fillType, false, sprayUsageVehicle.fillVolumeUnloadInfos[sprayUsageVehicle.newsprayer.unloadInfoIndex]);
        end
        hasSpray = oldFillLevel > 0;
    end
    if not hasSpray and self:getIsHired() and self.stopAiOnEmpty then
        local rootVehicle = self:getRootAttacherVehicle()
        rootVehicle:stopAIVehicle(AIVehicle.STOP_REASON_OUT_OF_FILL)
    end
    return hasSpray, fillType
end

function newSprayer:aiAllowedToRefill(superFunc)
    if self.needsTankActivation then
        local sprayUsageVehicle, _ = newSprayer.findAttachedSprayerTank(self:getRootAttacherVehicle(), self:getUnitFillTypes(self.newsprayer.fillUnitIndex), false, true);
        if sprayUsageVehicle == nil then
            return false;
        end;
    end;
    if superFunc ~= nil then
        return superFunc(self);
    end
    return true;
end;

function newSprayer:getCanAddHelpButtonText(superFunc)
    if self.activateOnLowering then
        return false;
    end;
    if superFunc ~= nil then
        return superFunc(self);
    end
    return true;
end;

function newSprayer.getTankActivators(currentVehicle, results)
    if currentVehicle.needsTankActivation then
        table.insert(results, currentVehicle);
    end
    for _,implement in pairs(currentVehicle.attachedImplements) do
        if implement.object ~= nil then
            Sprayer.getTankActivators(implement.object, results);
        end
    end
end

function newSprayer:getIsTurnedOnAllowed(superFunc, isTurnedOn)
    if not self.allowsSpraying then
        return false;
    end
    if self.activateOnLowering and self.isLowered ~= nil then
        return self:isLowered(false);
    end
    -- if sowingMachine is involved dismiss sprayer complains
    if isTurnedOn and self.sowingMachineGroundContactFlag ~= nil then
        return superFunc(self, isTurnedOn);
    end
    if isTurnedOn and self:getUnitFillLevel(self.newsprayer.fillUnitIndex) <= 0 then
        if self:getIsHired() then
            return true;
        end
        -- try to find tank
        local sprayerTank, fillType = newSprayer.findAttachedSprayerTank(self:getRootAttacherVehicle(), self:getUnitFillTypes(self.newsprayer.fillUnitIndex), self.needsTankActivation);
        if sprayerTank ~= nil and fillType ~= nil then
            return true;
        else
            return false;
        end
    end
    if superFunc ~= nil then
        return superFunc(self, isTurnedOn);
    end
    return true;
end

function newSprayer:onTurnedOn(noEventSend)
    if self.isClient and self.newSprayerSpecisDeactivated == false then
        if self.foldAnimTime == nil then
            -- If the sprayer is foldable, the spray valves are only turned on within the update loop
            for _, particleSystems in pairs(self.sprayParticleSystems) do
                for _, particleSystem in pairs(particleSystems) do
                    ParticleUtil.setEmittingState(particleSystem, false);
                end;
            end
            local fillType = self:getUnitLastValidFillType(self.newsprayer.fillUnitIndex);
            if fillType == FillUtil.FILLTYPE_UNKNOWN then
                fillType = self:getFirstEnabledFillType();
            end
            for _, sprayerEffect in pairs(self.newsprayerEffects) do
                EffectManager:setFillType(sprayerEffect.effects, fillType)
                EffectManager:startEffects(sprayerEffect.effects)
            end
        end
        if self.sprayingAnimationName ~= "" and self.playAnimation ~= nil then
            self:playAnimation(self.sprayingAnimationName, 1, self:getAnimationTime(self.sprayingAnimationName), true);
        end
    end
end

function newSprayer:onTurnedOff(noEventSend)
    if self.isClient and self.newSprayerSpecisDeactivated == false then
        for _, particleSystems in pairs(self.sprayParticleSystems) do
            for _, particleSystem in pairs(particleSystems) do
                ParticleUtil.setEmittingState(particleSystem, false);
            end
        end
        for _, sprayerEffect in pairs(self.newsprayerEffects) do
            EffectManager:stopEffects(sprayerEffect.effects);
        end
        SoundUtil.stopSample(self.sampleSprayer);
        if self.sprayingAnimationName ~= "" and self.stopAnimation ~= nil then
            self:stopAnimation(self.sprayingAnimationName, true);
        end
    end
end

function newSprayer:onAiTurnOn()
    if self.setIsTurnedOn ~= nil and self.newSprayerSpecisDeactivated == false then
        self:setIsTurnedOn(true, false);
    end
end

function newSprayer:onAiTurnOff()
    if self.setIsTurnedOn ~= nil and self.newSprayerSpecisDeactivated == false then
        self:setIsTurnedOn(false, false);
    end
end

function newSprayer:onAiLower()
    if self.setIsTurnedOn ~= nil and self.newSprayerSpecisDeactivated == false then
        self:setIsTurnedOn(true, false);
    end
end

function newSprayer:onAiRaise()
    if self.setIsTurnedOn ~= nil and self.newSprayerSpecisDeactivated == false then
        self:setIsTurnedOn(false, false);
    end
end

function newSprayer.findAttachedSprayerTank(currentVehicle, fillTypes, needsTankActivation, ignoreFillLevel)
    if currentVehicle.isSprayerTank and currentVehicle.getFillLevel ~= nil and currentVehicle.setFillLevel ~= nil then
        if (not needsTankActivation or (currentVehicle.getIsTurnedOn ~= nil and currentVehicle:getIsTurnedOn())) then
            local fillUnit = currentVehicle.fillUnits[currentVehicle.newsprayer.fillUnitIndex];
            for fillType,state in pairs(fillTypes) do
                if state and fillUnit.fillTypes[fillType] then
                    if (fillUnit.currentFillType == fillType and fillUnit.fillLevel > 0) or ignoreFillLevel then
                        return currentVehicle, fillType;
                    end
                end
            end
        end
    end
    for _,implement in pairs(currentVehicle.attachedImplements) do
        if implement.object ~= nil and implement.object.newSprayerSpecisDeactivated == false then
            local ret1, ret2 = newSprayer.findAttachedSprayerTank(implement.object, fillTypes, needsTankActivation, ignoreFillLevel);
            if ret1 ~= nil and ret2 ~= nil then
                return ret1, ret2;
            end
        end
    end
    return nil;
end

function newSprayer.getAktualFilltype(currentVehicle, fillTypes, needsTankActivation, ignoreFillLevel)
    if currentVehicle.isSprayerTank and currentVehicle.getFillLevel ~= nil and currentVehicle.setFillLevel ~= nil then
        if (not needsTankActivation or (currentVehicle.getIsTurnedOn ~= nil and currentVehicle:getIsTurnedOn())) then
            local fillUnit = currentVehicle.fillUnits[currentVehicle.newsprayer.fillUnitIndex];
            for fillType,state in pairs(fillTypes) do
                if state and fillUnit.fillTypes[fillType] then
                    if (fillUnit.currentFillType == fillType and fillUnit.fillLevel > 0) or ignoreFillLevel then
                        return currentVehicle, fillType;
                    end
                end
            end
        end
    end
end



function newSprayer:loadWorkAreaFromXML(superFunc, workArea, xmlFile, key)
    local retValue = true;
    if superFunc ~= nil then
        retValue = superFunc(self, workArea, xmlFile, key)
    end
    if workArea.type == WorkArea.AREATYPE_DEFAULT then
        workArea.type = WorkArea.AREATYPE_SPRAYER;
    end
    return retValue;
end

function newSprayer:getIsReadyToSpray()
    if self.lastSowingArea ~= nil then
        return self.lastSowingArea > 0;
    end
    return true;
end

function newSprayer:getAreEffectsVisible()
    return true;
end

function newSprayer:getnewSprayerLitersPerSecond(fillType)
    local scale = Utils.getNoNil(self.sprayUsageScale.fillTypeScales[fillType], self.sprayUsageScale.default)
    local litersPerSecond = 1
    local sprayTypeIndex = Sprayer.fillTypeToSprayType[fillType]
    if sprayTypeIndex ~= nil then
        local spray = Sprayer.sprayTypeIndexToDesc[sprayTypeIndex]
        if spray ~= nil then
            litersPerSecond = spray.litersPerSecond
        end
    end
--    return scale * litersPerSecond * self.speedLimit * self.sprayUsageScale.workingWidth
    return scale * litersPerSecond * (self.attacherVehicle.lastSpeed*self.attacherVehicle.speedDisplayScale*3600) * self.sprayUsageScalenewSprayer.workingWidth * self.newsprayAppRateMultiplyer
end

function newSprayer:doCheckSpeedLimit(superFunc)
    local parent = false;
    if superFunc ~= nil then
        parent = superFunc(self);
    end
    return parent or self.isSprayerSpeedLimitActive;
end

function newSprayer.getDefaultSpeedLimit()
    return 15;
end

function newSprayer:processSprayerAreas(workAreas, fillType)
	local isFieldJobVehicle = false;
	if g_currentMission.fieldJobManager ~= nil then
		if g_currentMission.fieldJobManager.currentFieldJob ~= nil then
			if g_currentMission.fieldJobManager.currentFieldJob.fieldJobVehicles ~= nil then
				for _,vehicle in pairs(g_currentMission.fieldJobManager.currentFieldJob.fieldJobVehicles) do
					if vehicle == self then
						isFieldJobVehicle = true;
						break;
					end;
				end;
			end;
		end;
	end;

    local sprayType = 1;
    if (fillType == FillUtil.FILLTYPE_MANURE or fillType == FillUtil.FILLTYPE_LIQUIDMANURE or fillType == FillUtil.FILLTYPE_DIGESTATE) and isFieldJobVehicle == false and self.newsprayerTextureDisable == false then 
        sprayType = 2;
    end
    local totalPixels = 0;
    local numAreas = table.getn(workAreas);
    for i=1, numAreas do
        local x = workAreas[i][1];
        local z = workAreas[i][2];
        local x1 = workAreas[i][3];
        local z1 = workAreas[i][4];
        local x2 = workAreas[i][5];
        local z2 = workAreas[i][6];
        local pixels, pixelsTotal;        
        if self.cultivatorGroundContactFlag ~= nil and self.sowingMachineGroundContactFlag == nil then
            pixels, pixelsTotal = Utils.updateSprayArea(x, z, x1, z1, x2, z2, g_currentMission.cultivatorValue, sprayType);
        else
            pixels, pixelsTotal = Utils.updateSprayArea(x, z, x1, z1, x2, z2, nil, sprayType);
        end
        totalPixels = totalPixels + pixels;
    end;
    return totalPixels;
end;

function newSprayer:getSaveAttributesAndNodes(nodeIdent)
end

function newSprayer:mouseEvent(posX, posY, isDown, isUp, button)
end;

function newSprayer:keyEvent(unicode, sym, modifier, isDown)
end;

function newSprayer:onAttach()
end; 	

function newSprayer:onDetach()
end;

--Events fuer Dediserver--
--Test fuer Dediserver--
newSprayer_Sprayrate_Event = {};
newSprayer_Sprayrate_Event_mt = Class(newSprayer_Sprayrate_Event, Event);
  
InitEventClass(newSprayer_Sprayrate_Event, "newSprayer_Sprayrate_Event");
  
function newSprayer_Sprayrate_Event:emptyNew()
    local self = Event:new(newSprayer_Sprayrate_Event_mt);
    return self;
end;
  
function newSprayer_Sprayrate_Event:new(object, data)
    local self = newSprayer_Sprayrate_Event:emptyNew()
    self.data = data;
  
    self.object = object;
    return self;
end;
  
function newSprayer_Sprayrate_Event:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.data = streamReadFloat32(streamId);
      
	self.object = networkGetObject(id);
    self:run(connection);
end;
  
function newSprayer_Sprayrate_Event:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.object));
    streamWriteFloat32(streamId, self.data);
end;
  
function newSprayer_Sprayrate_Event:run(connection)
    self.object.newsprayAppRateMultiplyer = self.data;
	  	  
    if not connection:getIsServer() then
        g_server:broadcastEvent(newSprayer_Sprayrate_Event:new(self.object, self.data), nil, connection, self.object);
    end;
end;
 
--Test fuer Dediserver--
newSprayer_TextureDisable_Event = {};
newSprayer_TextureDisable_Event_mt = Class(newSprayer_TextureDisable_Event, Event);
  
InitEventClass(newSprayer_TextureDisable_Event, "newSprayer_TextureDisable_Event");
  
function newSprayer_TextureDisable_Event:emptyNew()
    local self = Event:new(newSprayer_TextureDisable_Event_mt);
    return self;
end;
  
function newSprayer_TextureDisable_Event:new(object, data)
    local self = newSprayer_TextureDisable_Event:emptyNew()
    self.data = data;
  
    self.object = object;
    return self;
end;
  
function newSprayer_TextureDisable_Event:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.data = streamReadBool(streamId);
      
	self.object = networkGetObject(id);
    self:run(connection);
end;
  
function newSprayer_TextureDisable_Event:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.object));
    streamWriteBool(streamId, self.data);
end;
  
function newSprayer_TextureDisable_Event:run(connection)
    self.object.newsprayerTextureDisable = self.data;
	  	  
    if not connection:getIsServer() then
        g_server:broadcastEvent(newSprayer_TextureDisable_Event:new(self.object, self.data), nil, connection, self.object);
    end;
end;
  
